import lenstronomy.Sampling.Likelihoods.image_likelihood as img_likelihood
import numpy as np
import pytest


class TestImageLikelihood(object):
    def setup_method(self):
        pass

    def test_create_im_sim(self):
        pass


if __name__ == "__main__":
    pytest.main()
