__author__ = "lenstronomy developers"
__email__ = "lenstronomy-dev@googlegroups.com"
__version__ = "1.11.5"
__credits__ = "ETH Zurich, UCLA, Stanford, Stony Brook"

from .Util.package_util import short, laconic
